@@ -1,6 +1,6 @@
 /* img.c
  * Routines for reading and writing Survex ".3d" image files
- * Copyright (C) 1993-2003 Olly Betts
+ * Copyright (C) 1993-2004,2005,2006,2010 Olly Betts
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -26,9 +26,18 @@
 #include <stdlib.h>
 #include <time.h>
 #include <ctype.h>
+#include <locale.h>
 
 #include "img.h"
 
+#define LATEST_IMG_VERSION 7
+
+#ifndef IMG_API_VERSION
+# define IMG_API_VERSION 0
+#elif IMG_API_VERSION > 1
+# error IMG_API_VERSION > 1 too new
+#endif
+
 #ifdef IMG_HOSTED
 # include "debug.h"
 # include "filelist.h"
@@ -37,8 +46,9 @@
 # include "useful.h"
 # define TIMENA msg(/*Date and time not available.*/108)
 # define TIMEFMT msg(/*%a,%Y.%m.%d %H:%M:%S %Z*/107)
+# define INT32_T int32_t
 #else
-# define INT32_T long
+# define INT32_T int
 # define TIMENA "Time not available."
 # define TIMEFMT "%a,%Y.%m.%d %H:%M:%S %Z"
 # define EXT_SVX_3D "3d"
@@ -54,6 +64,8 @@
 /* open file FNM with mode MODE, maybe using path PTH and/or extension EXT */
 /* path isn't used in img.c, but EXT is */
 # define fopenWithPthAndExt(PTH,FNM,EXT,MODE,X) fopen(FNM,MODE)
+# define PUTC(C, FH) putc(C, FH)
+# define GETC(FH) getc(FH)
 # define fFalse 0
 # define fTrue 1
 # define bool int
@@ -61,41 +73,54 @@
  * if (s) fputsnl(s,fh); else exit(1);
  * to work as intended
  */
-# define fputsnl(S, FH) do {fputs((S), (FH)); putc('\n', (FH));} while(0)
+# define fputsnl(S, FH) do {fputs((S), (FH)); PUTC('\n', (FH));} while(0)
 # define SVX_ASSERT(X)
+/* In C++ code, #include<algorithm> and use std::max and std::min instead. */
+#ifndef __cplusplus
+/* Return max/min of two numbers. */
+/* May be defined already (e.g. by Borland C in stdlib.h) */
+/* NB Bad news if X or Y has side-effects... */
+# ifndef max
+#  define max(X, Y) ((X) > (Y) ? (X) : (Y))
+# endif
+# ifndef min
+#  define min(X, Y) ((X) < (Y) ? (X) : (Y))
+# endif
+#endif
+
+static INT32_T
 
-static long
 get32(FILE *fh)
 {
-   long w = getc(fh);
-   w |= (long)getc(fh) << 8l;
-   w |= (long)getc(fh) << 16l;
-   w |= (long)getc(fh) << 24l;
+   INT32_T w = GETC(fh);
+   w |= (INT32_T)GETC(fh) << 8l;
+   w |= (INT32_T)GETC(fh) << 16l;
+   w |= (INT32_T)GETC(fh) << 24l;
    return w;
 }
 
 static void
 put32(long w, FILE *fh)
 {
-   putc((char)(w), fh);
-   putc((char)(w >> 8l), fh);
-   putc((char)(w >> 16l), fh);
-   putc((char)(w >> 24l), fh);
+   PUTC((char)(w), fh);
+   PUTC((char)(w >> 8l), fh);
+   PUTC((char)(w >> 16l), fh);
+   PUTC((char)(w >> 24l), fh);
 }
 
 static short
 get16(FILE *fh)
 {
-   short w = getc(fh);
-   w |= (long)getc(fh) << 8l;
+   short w = GETC(fh);
+   w |= (short)GETC(fh) << 8l;
    return w;
 }
 
 static void
 put16(short w, FILE *fh)
 {
-   putc((char)(w), fh);
-   putc((char)(w >> 8l), fh);
+   PUTC((char)(w), fh);
+   PUTC((char)(w >> 8l), fh);
 }
 
 static char *
@@ -122,6 +147,12 @@ baseleaf_from_fnm(const char *fnm)
 }
 #endif
 
+static unsigned short
+getu16(FILE *fh)
+{
+   return (unsigned short)get16(fh);
+}
+
 #include <math.h>
 #ifdef HAVE_ROUND
 extern double round(double); /* prototype is often missing... */
@@ -150,7 +181,7 @@ int my_strcasecmp(const char *s1, const char *s2) {
 }
 #endif
 
-unsigned int img_output_version = 3;
+unsigned int img_output_version = LATEST_IMG_VERSION;
 
 #ifdef IMG_HOSTED
 static enum {
@@ -177,6 +208,12 @@ static img_errcode img_errno = IMG_NONE;
 /* Attempt to string paste to ensure we are passed a literal string */
 #define LITLEN(S) (sizeof(S"") - 1)
 
+/* Fake "version numbers" for non-3d formats we can read. */
+#define VERSION_CMAP_SHOT	-4
+#define VERSION_CMAP_STATION	-3
+#define VERSION_COMPASS_PLT	-2
+#define VERSION_SURVEX_POS	-1
+
 static char *
 my_strdup(const char *str)
 {
@@ -196,7 +233,7 @@ getline_alloc(FILE *fh)
    char *buf = xosmalloc(len);
    if (!buf) return NULL;
 
-   ch = getc(fh);
+   ch = GETC(fh);
    while (ch != '\n' && ch != '\r' && ch != EOF) {
       buf[i++] = ch;
       if (i == len - 1) {
@@ -209,11 +246,11 @@ getline_alloc(FILE *fh)
 	 }
 	 buf = p;
       }
-      ch = getc(fh);
+      ch = GETC(fh);
    }
    if (ch == '\n' || ch == '\r') {
       int otherone = ch ^ ('\n' ^ '\r');
-      ch = getc(fh);
+      ch = GETC(fh);
       /* if it's not the other eol character, put it back */
       if (ch != otherone) ungetc(ch, fh);
    }
@@ -279,7 +316,7 @@ img_open_survey(const char *fnm, const char *survey)
       return NULL;
    }
 
-   pimg->fh = fopenWithPthAndExt("", fnm, EXT_SVX_3D, "rb", NULL);
+   pimg->fh = fopenWithPthAndExt("", fnm, EXT_SVX_3D, "rb", &(pimg->filename_opened));
    if (pimg->fh == NULL) {
       osfree(pimg->label_buf);
       osfree(pimg);
@@ -292,16 +329,25 @@ img_open_survey(const char *fnm, const char *survey)
 
    pimg->flags = 0;
 
-   /* for version 3 we use label_buf to store the prefix for reuse */
-   /* for version -2, 0 value indicates we haven't entered a survey yet */
-   /* for version -4, we store the last station here to detect whether
-    * we MOVE or LINE */
+   /* for version >= 3 we use label_buf to store the prefix for reuse */
+   /* for VERSION_COMPASS_PLT, 0 value indicates we haven't
+    * entered a survey yet */
+   /* for VERSION_CMAP_SHOT, we store the last station here
+    * to detect whether we MOVE or LINE */
    pimg->label_len = 0;
    pimg->label_buf[0] = '\0';
 
    pimg->survey = NULL;
    pimg->survey_len = 0;
    pimg->separator = '.';
+#if IMG_API_VERSION == 0
+   pimg->date1 = pimg->date2 = 0;
+#else /* IMG_API_VERSION == 1 */
+   pimg->days1 = pimg->days2 = -1;
+#endif
+   pimg->is_extended_elevation = 0;
+
+   pimg->l = pimg->r = pimg->u = pimg->d = -1.0;
 
    pimg->title = pimg->datestamp = NULL;
    if (survey) {
@@ -332,17 +378,18 @@ img_open_survey(const char *fnm, const char *survey)
       pimg->survey_len = len;
    }
 
-   /* [version -2, -3, -4] pending IMG_LINE or IMG_MOVE - both have 4 added
-    * [version -1] already skipped heading line, or there wasn't one
+   /* [VERSION_COMPASS_PLT, VERSION_CMAP_STATION, VERSION_CMAP_SHOT] pending
+    * IMG_LINE or IMG_MOVE - both have 4 added.
+    * [VERSION_SURVEX_POS] already skipped heading line, or there wasn't one
     * [version 0] not in the middle of a 'LINE' command
-    * [version 3] not in the middle of turning a LINE into a MOVE
+    * [version >= 3] not in the middle of turning a LINE into a MOVE
     */
    pimg->pending = 0;
 
    len = strlen(fnm);
    if (has_ext(fnm, len, EXT_SVX_POS)) {
 pos_file:
-      pimg->version = -1;
+      pimg->version = VERSION_SURVEX_POS;
       if (!pimg->survey) pimg->title = baseleaf_from_fnm(fnm);
       pimg->datestamp = my_strdup(TIMENA);
       if (!pimg->datestamp) {
@@ -356,7 +403,7 @@ pos_file:
    if (has_ext(fnm, len, EXT_PLT) || has_ext(fnm, len, EXT_PLF)) {
       long fpos;
 plt_file:
-      pimg->version = -2;
+      pimg->version = VERSION_COMPASS_PLT;
       /* Spaces aren't legal in Compass station names, but dots are, so
        * use space as the level separator */
       pimg->separator = ' ';
@@ -368,7 +415,7 @@ plt_file:
 	 goto error;
       }
       while (1) {
-	 ch = getc(pimg->fh);
+	 ch = GETC(pimg->fh);
 	 switch (ch) {
 	  case '\x1a':
 	    fseek(pimg->fh, -1, SEEK_CUR);
@@ -420,7 +467,7 @@ plt_file:
 	    break;
 	 }
 	 while (ch != '\n' && ch != '\r') {
-	    ch = getc(pimg->fh);
+	    ch = GETC(pimg->fh);
 	 }
       }
    }
@@ -429,7 +476,7 @@ plt_file:
       char *line;
 xyz_file:
       /* Spaces aren't legal in CMAP station names, but dots are, so
-       * use space as the level separator */
+       * use space as the level separator. */
       pimg->separator = ' ';
       line = getline_alloc(pimg->fh);
       if (!line) {
@@ -474,9 +521,9 @@ xyz_file:
 	 goto error;
       }
       if (line[1] == 'S') {
-	 pimg->version = -3; /* Station format */
+	 pimg->version = VERSION_CMAP_STATION;
       } else {
-	 pimg->version = -4; /* Shot format */
+	 pimg->version = VERSION_CMAP_SHOT;
       }
       osfree(line);
       line = getline_alloc(pimg->fh);
@@ -521,18 +568,18 @@ xyz_file:
    }
 
    /* check file format version */
-   ch = getc(pimg->fh);
+   ch = GETC(pimg->fh);
    pimg->version = 0;
    if (tolower(ch) == 'b') {
       /* binary file iff B/b prefix */
       pimg->version = 1;
-      ch = getc(pimg->fh);
+      ch = GETC(pimg->fh);
    }
    if (ch != 'v') {
       img_errno = IMG_BADFORMAT;
       goto error;
    }
-   ch = getc(pimg->fh);
+   ch = GETC(pimg->fh);
    if (ch == '0') {
       if (fread(buf, 4, 1, pimg->fh) != 1 || memcmp(buf, ".01\n", 4) != 0) {
 	 img_errno = IMG_BADFORMAT;
@@ -540,7 +587,7 @@ xyz_file:
       }
       /* nothing special to do */
    } else if (pimg->version == 0) {
-      if (ch < '2' || ch > '3' || getc(pimg->fh) != '\n') {
+      if (ch < '2' || ch > '0' + LATEST_IMG_VERSION || GETC(pimg->fh) != '\n') {
 	 img_errno = IMG_TOONEW;
 	 goto error;
       }
@@ -561,6 +608,9 @@ v03d:
       error:
       osfree(pimg->title);
       osfree(pimg->datestamp);
+#ifdef IMG_HOSTED
+      osfree(pimg->filename_opened);
+#endif
       fclose(pimg->fh);
       osfree(pimg);
       return NULL;
@@ -568,6 +618,12 @@ v03d:
 
    pimg->start = ftell(pimg->fh);
 
+   len = strlen(pimg->title);
+   if (len > 11 && strcmp(pimg->title + len - 11, " (extended)") == 0) {
+       pimg->title[len - 11] = '\0';
+       pimg->is_extended_elevation = 1;
+   }
+
    return pimg;
 }
 
@@ -583,16 +639,17 @@ img_rewind(img *pimg)
       return 0;
    }
    clearerr(pimg->fh);
-   /* [version -1] already skipped heading line, or there wasn't one
+   /* [VERSION_SURVEX_POS] already skipped heading line, or there wasn't one
     * [version 0] not in the middle of a 'LINE' command
-    * [version 3] not in the middle of turning a LINE into a MOVE */
+    * [version >= 3] not in the middle of turning a LINE into a MOVE */
    pimg->pending = 0;
 
    img_errno = IMG_NONE;
 
-   /* for version 3 we use label_buf to store the prefix for reuse */
-   /* for version -2, 0 value indicates we haven't entered a survey yet */
-   /* for version -4, we store the last station here to detect whether
+   /* for version >= 3 we use label_buf to store the prefix for reuse */
+   /* for VERSION_COMPASS_PLT, 0 value indicates we haven't entered a survey
+    * yet */
+   /* for VERSION_CMAP_SHOT, we store the last station here to detect whether
     * we MOVE or LINE */
    pimg->label_len = 0;
    return 1;
@@ -633,13 +690,15 @@ img_open_write(const char *fnm, char *title_buf, bool fBinary)
       return NULL;
    }
 
+   pimg->filename_opened = NULL;
+
    /* Output image file header */
    fputs("Survex 3D Image File\n", pimg->fh); /* file identifier string */
    if (img_output_version < 2) {
       pimg->version = 1;
       fputs("Bv0.01\n", pimg->fh); /* binary file format version number */
    } else {
-      pimg->version = (img_output_version > 2) ? 3 : 2;
+      pimg->version = (img_output_version > LATEST_IMG_VERSION) ? LATEST_IMG_VERSION : img_output_version;
       fprintf(pimg->fh, "v%d\n", pimg->version); /* file format version no. */
    }
    fputsnl(title_buf, pimg->fh);
@@ -652,13 +711,36 @@ img_open_write(const char *fnm, char *title_buf, bool fBinary)
       strftime(date, 256, TIMEFMT, localtime(&tm));
       fputsnl(date, pimg->fh);
    }
+#if 0
+   if (img_output_version >= 5) {
+       static const unsigned char codelengths[32] = {
+	   4,  8,  8,  16, 0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
+	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
+       };
+       fwrite(codelengths, 32, 1, pimg->fh);
+   }
+#endif
    pimg->fRead = fFalse; /* writing to this file */
    img_errno = IMG_NONE;
 
-   /* for version 3 we use label_buf to store the prefix for reuse */
+   /* for version >= 3 we use label_buf to store the prefix for reuse */
    pimg->label_buf[0] = '\0';
    pimg->label_len = 0;
 
+#if IMG_API_VERSION == 0
+   pimg->date1 = pimg->date2 = 0;
+   pimg->olddate1 = pimg->olddate2 = 0;
+#else /* IMG_API_VERSION == 1 */
+   pimg->days1 = pimg->days2 = -1;
+   pimg->olddays1 = pimg->olddays2 = -1;
+#endif
+
+   pimg->l = pimg->r = pimg->u = pimg->d = -1.0;
+
+   pimg->n_legs = 0;
+   pimg->length = 0.0;
+   pimg->E = pimg->H = pimg->V = 0.0;
+
    /* Don't check for write errors now - let img_close() report them... */
    return pimg;
 }
@@ -726,23 +808,75 @@ skip_coord(FILE *fh)
     return (fseek(fh, 12, SEEK_CUR) == 0);
 }
 
+static int
+read_v3label(img *pimg)
+{
+   char *q;
+   long len = GETC(pimg->fh);
+   if (len == EOF) {
+      img_errno = feof(pimg->fh) ? IMG_BADFORMAT : IMG_READERROR;
+      return img_BAD;
+   }
+   if (len == 0xfe) {
+      len += get16(pimg->fh);
+      if (feof(pimg->fh)) {
+	 img_errno = IMG_BADFORMAT;
+	 return img_BAD;
+      }
+      if (ferror(pimg->fh)) {
+	 img_errno = IMG_READERROR;
+	 return img_BAD;
+      }
+   } else if (len == 0xff) {
+      len = get32(pimg->fh);
+      if (ferror(pimg->fh)) {
+	 img_errno = IMG_READERROR;
+	 return img_BAD;
+      }
+      if (feof(pimg->fh) || len < 0xfe + 0xffff) {
+	 img_errno = IMG_BADFORMAT;
+	 return img_BAD;
+      }
+   }
+
+   if (!check_label_space(pimg, pimg->label_len + len + 1)) {
+      img_errno = IMG_OUTOFMEMORY;
+      return img_BAD;
+   }
+   q = pimg->label_buf + pimg->label_len;
+   pimg->label_len += len;
+   if (len && fread(q, len, 1, pimg->fh) != 1) {
+      img_errno = feof(pimg->fh) ? IMG_BADFORMAT : IMG_READERROR;
+      return img_BAD;
+   }
+   q[len] = '\0';
+   return 0;
+}
+
+static int img_read_item_ascii(img *pimg, img_point *p);
+
 int
 img_read_item(img *pimg, img_point *p)
 {
    int result;
    pimg->flags = 0;
 
-   if (pimg->version == 3) {
+   if (pimg->version >= 3) {
       int opt;
+      pimg->l = pimg->r = pimg->u = pimg->d = -1.0;
+      if (pimg->pending == 256) {
+	 pimg->pending = 0;
+	 return img_XSECT_END;
+      }
       if (pimg->pending >= 0x80) {
 	 *p = pimg->mv;
 	 pimg->flags = (int)(pimg->pending) & 0x3f;
 	 pimg->pending = 0;
 	 return img_LINE;
       }
-      again3: /* label to goto if we get a prefix */
+      again3: /* label to goto if we get a prefix, date, or lrud */
       pimg->label = pimg->label_buf;
-      opt = getc(pimg->fh);
+      opt = GETC(pimg->fh);
       if (opt == EOF) {
 	 img_errno = feof(pimg->fh) ? IMG_BADFORMAT : IMG_READERROR;
 	 return img_BAD;
@@ -757,12 +891,13 @@ img_read_item(img *pimg, img_point *p)
 	 if (opt < 15) {
 	    /* 1-14 mean trim that many levels from current prefix */
 	    int c;
-	    if (pimg->label_len <= 16) {
+	    if (pimg->label_len <= 17) {
 	       /* zero prefix using "0" */
 	       img_errno = IMG_BADFORMAT;
 	       return img_BAD;
 	    }
-	    c = pimg->label_len - 16 - 1;
+	    /* extra - 1 because label_len points to one past the end */
+	    c = pimg->label_len - 17 - 1;
 	    while (pimg->label_buf[c] != '.' || --opt > 0) {
 	       if (--c < 0) {
 		  /* zero prefix using "0" */
@@ -778,6 +913,121 @@ img_read_item(img *pimg, img_point *p)
 	    result = img_MOVE;
 	    break;
 	 }
+	 if (opt >= 0x20) {
+	     switch (opt) {
+		 case 0x20: /* Single date */
+		     if (pimg->version < 7) {
+			 int date1 = get32(pimg->fh);
+#if IMG_API_VERSION == 0
+			 pimg->date2 = pimg->date1 = date1;
+#else /* IMG_API_VERSION == 1 */
+			 if (date1 != 0) {
+			     pimg->days2 = pimg->days1 = (date1 / 86400) + 25567;
+			 } else {
+			     pimg->days2 = pimg->days1 = -1;
+			 }
+#endif
+		     } else {
+			 int days1 = (int)getu16(pimg->fh);
+#if IMG_API_VERSION == 0
+			 pimg->date2 = pimg->date1 = (days1 - 25567) * 86400;
+#else /* IMG_API_VERSION == 1 */
+			 pimg->days2 = pimg->days1 = days1;
+#endif
+		     }
+		     break;
+		 case 0x21: /* Date range (short for v7+) */
+		     if (pimg->version < 7) {
+			 INT32_T date1 = get32(pimg->fh);
+			 INT32_T date2 = get32(pimg->fh);
+#if IMG_API_VERSION == 0
+			 pimg->date1 = date1;
+			 pimg->date2 = date2;
+#else /* IMG_API_VERSION == 1 */
+			 pimg->days1 = (date1 / 86400) + 25567;
+			 pimg->days2 = (date2 / 86400) + 25567;
+#endif
+		     } else {
+			 int days1 = (int)getu16(pimg->fh);
+			 int days2 = days1 + GETC(pimg->fh) + 1;
+#if IMG_API_VERSION == 0
+			 pimg->date1 = (pimg->days1 - 25567) * 86400;
+			 pimg->date2 = (pimg->days2 - 25567) * 86400;
+#else /* IMG_API_VERSION == 1 */
+			 pimg->days1 = days1;
+			 pimg->days2 = days2;
+#endif
+		     }
+		     break;
+		 case 0x22: /* Error info */
+		     pimg->n_legs = get32(pimg->fh);
+		     pimg->length = get32(pimg->fh) / 100.0;
+		     pimg->E = get32(pimg->fh) / 100.0;
+		     pimg->H = get32(pimg->fh) / 100.0;
+		     pimg->V = get32(pimg->fh) / 100.0;
+		     return img_ERROR_INFO;
+		 case 0x23: { /* v7+: Date range (long) */
+		     if (pimg->version < 7) {
+			 img_errno = IMG_BADFORMAT;
+			 return img_BAD;
+		     }
+		     int days1 = (int)getu16(pimg->fh);
+		     int days2 = (int)getu16(pimg->fh);
+#if IMG_API_VERSION == 0
+		     pimg->date1 = (days1 - 25567) * 86400;
+		     pimg->date2 = (days2 - 25567) * 86400;
+#else /* IMG_API_VERSION == 1 */
+		     pimg->days1 = days1;
+		     pimg->days2 = days2;
+#endif
+		     break;
+		 }
+		 case 0x24: { /* v7+: No date info */
+#if IMG_API_VERSION == 0
+		     pimg->date1 = pimg->date2 = 0;
+#else /* IMG_API_VERSION == 1 */
+		     pimg->days1 = pimg->days2 = -1;
+#endif
+		     break;
+		 }
+		 case 0x30: case 0x31: /* LRUD */
+		 case 0x32: case 0x33: /* Big LRUD! */
+		     if (read_v3label(pimg) == img_BAD) return img_BAD;
+		     pimg->flags = (int)opt & 0x01;
+		     if (opt < 0x32) {
+			 pimg->l = get16(pimg->fh) / 100.0;
+			 pimg->r = get16(pimg->fh) / 100.0;
+			 pimg->u = get16(pimg->fh) / 100.0;
+			 pimg->d = get16(pimg->fh) / 100.0;
+		     } else {
+			 pimg->l = get32(pimg->fh) / 100.0;
+			 pimg->r = get32(pimg->fh) / 100.0;
+			 pimg->u = get32(pimg->fh) / 100.0;
+			 pimg->d = get32(pimg->fh) / 100.0;
+		     }
+		     if (pimg->survey_len) {
+			 size_t l = pimg->survey_len;
+			 const char *s = pimg->label_buf;
+			 if (strncmp(pimg->survey, s, l + 1) != 0) {
+			     return img_XSECT_END;
+			 }
+			 pimg->label += l;
+			 /* skip the dot if there */
+			 if (*pimg->label) pimg->label++;
+		     }
+		     /* If this is the last cross-section in this passage, set
+		      * pending so we return img_XSECT_END next time. */
+		     if (pimg->flags & 0x01) {
+			 pimg->pending = 256;
+			 pimg->flags &= ~0x01;
+		     }
+		     return img_XSECT;
+		 default: /* 0x25 - 0x2f and 0x34 - 0x3f are currently unallocated. */
+		     img_errno = IMG_BADFORMAT;
+		     return img_BAD;
+	     }
+	     goto again3;
+	 }
 	 /* 16-31 mean remove (n - 15) characters from the prefix */
 	 /* zero prefix using 0 */
 	 if (pimg->label_len <= (size_t)(opt - 15)) {
@@ -786,72 +1036,46 @@ img_read_item(img *pimg, img_point *p)
 	 }
 	 pimg->label_len -= (opt - 15);
 	 goto again3;
-       case 1: case 2: {
-	 char *q;
-	 long len = getc(pimg->fh);
-	 if (len == EOF) {
-	    img_errno = feof(pimg->fh) ? IMG_BADFORMAT : IMG_READERROR;
-	    return img_BAD;
-	 }
-	 if (len == 0xfe) {
-	    len += get16(pimg->fh);
-	    if (feof(pimg->fh)) {
-	       img_errno = IMG_BADFORMAT;
-	       return img_BAD;
-	    }
-	    if (ferror(pimg->fh)) {
-	       img_errno = IMG_READERROR;
-	       return img_BAD;
-	    }
-	 } else if (len == 0xff) {
-	    len = get32(pimg->fh);
-	    if (ferror(pimg->fh)) {
-	       img_errno = IMG_READERROR;
-	       return img_BAD;
-	    }
-	    if (feof(pimg->fh) || len < 0xfe + 0xffff) {
-	       img_errno = IMG_BADFORMAT;
-	       return img_BAD;
-	    }
-	 }
+       case 1:
+	 if (read_v3label(pimg) == img_BAD) return img_BAD;
 
-	 if (!check_label_space(pimg, pimg->label_len + len + 1)) {
-	    img_errno = IMG_OUTOFMEMORY;
-	    return img_BAD;
-	 }
-	 q = pimg->label_buf + pimg->label_len;
-	 pimg->label_len += len;
-	 if (len && fread(q, len, 1, pimg->fh) != 1) {
-	    img_errno = feof(pimg->fh) ? IMG_BADFORMAT : IMG_READERROR;
-	    return img_BAD;
-	 }
-	 q[len] = '\0';
-
-	 result = opt & 0x40 ? img_LABEL : img_LINE;
+	 result = img_LABEL;
 
 	 if (pimg->survey_len) {
 	    size_t l = pimg->survey_len;
 	    const char *s = pimg->label_buf;
-	    if (result == img_LINE) {
-	       if (strncmp(pimg->survey, s, l) != 0 ||
-		   !(s[l] == '.' || s[l] == '\0')) {
-		  if (!read_coord(pimg->fh, &(pimg->mv))) return img_BAD;
-		  pimg->pending = 15;
-		  goto again3;
-	       }
-	    } else {
-	       if (strncmp(pimg->survey, s, l + 1) != 0) {
-		  if (!skip_coord(pimg->fh)) return img_BAD;
-		  pimg->pending = 0;
-		  goto again3;
-	       }
+	    if (strncmp(pimg->survey, s, l + 1) != 0) {
+	       if (!skip_coord(pimg->fh)) return img_BAD;
+	       pimg->pending = 0;
+	       goto again3;
 	    }
 	    pimg->label += l;
 	    /* skip the dot if there */
 	    if (*pimg->label) pimg->label++;
 	 }
 
-	 if (result == img_LINE && pimg->pending) {
+	 pimg->flags = (int)opt & 0x3f;
+	 break;
+       case 2:
+	 if (read_v3label(pimg) == img_BAD) return img_BAD;
+
+	 result = img_LINE;
+
+	 if (pimg->survey_len) {
+	    size_t l = pimg->survey_len;
+	    const char *s = pimg->label_buf;
+	    if (strncmp(pimg->survey, s, l) != 0 ||
+		!(s[l] == '.' || s[l] == '\0')) {
+	       if (!read_coord(pimg->fh, &(pimg->mv))) return img_BAD;
+	       pimg->pending = 15;
+	       goto again3;
+	    }
+	    pimg->label += l;
+	    /* skip the dot if there */
+	    if (*pimg->label) pimg->label++;
+	 }
+
+	 if (pimg->pending) {
 	    *p = pimg->mv;
 	    if (!read_coord(pimg->fh, &(pimg->mv))) return img_BAD;
 	    pimg->pending = opt;
@@ -859,7 +1083,6 @@ img_read_item(img *pimg, img_point *p)
 	 }
 	 pimg->flags = (int)opt & 0x3f;
 	 break;
-       }
        default:
 	 img_errno = IMG_BADFORMAT;
 	 return img_BAD;
@@ -886,7 +1109,7 @@ img_read_item(img *pimg, img_point *p)
 	    opt = get32(pimg->fh);
 	 }
       } else {
-	 opt = getc(pimg->fh);
+	 opt = GETC(pimg->fh);
       }
 
       if (feof(pimg->fh)) {
@@ -912,7 +1135,7 @@ img_read_item(img *pimg, img_point *p)
 	 char *q;
 	 int ch;
 	 result = img_LABEL;
-	 ch = getc(pimg->fh);
+	 ch = GETC(pimg->fh);
 	 if (ch == EOF) {
 	    img_errno = feof(pimg->fh) ? IMG_BADFORMAT : IMG_READERROR;
 	    return img_BAD;
@@ -944,7 +1167,7 @@ img_read_item(img *pimg, img_point *p)
 	 result = img_LABEL;
 
 	 if (opt == 7)
-	    pimg->flags = getc(pimg->fh);
+	    pimg->flags = GETC(pimg->fh);
 	 else
 	    pimg->flags = img_SFLAG_UNDERGROUND; /* no flags given... */
 
@@ -1039,7 +1262,25 @@ img_read_item(img *pimg, img_point *p)
       }
 
       return result;
-   } else if (pimg->version == 0) {
+   } else {
+      /* We need to set the default locale for fscanf() to work on
+       * numbers with "." as decimal point. */
+      char * current_locale = my_strdup(setlocale(LC_NUMERIC, NULL));
+      setlocale(LC_NUMERIC, "C");
+      result = img_read_item_ascii(pimg, p);
+      setlocale(LC_NUMERIC, current_locale);
+      free(current_locale);
+      return result;
+   }
+}
+
+/* Handle all for ASCII formats so we can easily fix up the locale
+ * so fscanf(), etc work on numbers with "." as decimal point. */
+static int
+img_read_item_ascii(img *pimg, img_point *p)
+{
+   int result;
+   if (pimg->version == 0) {
       ascii_again:
       pimg->label[0] = '\0';
       if (feof(pimg->fh)) return img_STOP;
@@ -1066,8 +1307,8 @@ img_read_item(img *pimg, img_point *p)
 	    goto ascii_again;
 	 } else if (strcmp(cmd, "name") == 0) {
 	    size_t off = 0;
-	    int ch = getc(pimg->fh);
-	    if (ch == ' ') ch = getc(pimg->fh);
+	    int ch = GETC(pimg->fh);
+	    if (ch == ' ') ch = GETC(pimg->fh);
 	    while (ch != ' ') {
 	       if (ch == '\n' || ch == EOF) {
 		  img_errno = ferror(pimg->fh) ? IMG_READERROR : IMG_BADFORMAT;
@@ -1080,13 +1321,15 @@ img_read_item(img *pimg, img_point *p)
 		  }
 	       }
 	       pimg->label_buf[off++] = ch;
-	       ch = getc(pimg->fh);
+	       ch = GETC(pimg->fh);
 	    }
 	    pimg->label_buf[off] = '\0';
 
 	    pimg->label = pimg->label_buf;
 	    if (pimg->label[0] == '\\') pimg->label++;
 
+	    pimg->flags = img_SFLAG_UNDERGROUND; /* default flags */
+
 	    result = img_LABEL;
 	 } else {
 	    img_errno = IMG_BADFORMAT;
@@ -1106,8 +1349,8 @@ img_read_item(img *pimg, img_point *p)
       }
 
       return result;
-   } else if (pimg->version == -1) {
-      /* version -1: .pos file */
+   } else if (pimg->version == VERSION_SURVEX_POS) {
+      /* Survex .pos file */
       size_t off;
       pimg->flags = img_SFLAG_UNDERGROUND; /* default flags */
       againpos:
@@ -1126,7 +1369,7 @@ img_read_item(img *pimg, img_point *p)
 	 pimg->pending = 1;
 	 /* ignore rest of line */
 	 do {
-	    ch = getc(pimg->fh);
+	    ch = GETC(pimg->fh);
 	 } while (ch != '\n' && ch != '\r' && ch != EOF);
       }
 
@@ -1159,8 +1402,8 @@ img_read_item(img *pimg, img_point *p)
       }
 
       return img_LABEL;
-   } else if (pimg->version == -2) {
-      /* version -2: Compass .plt file */
+   } else if (pimg->version == VERSION_COMPASS_PLT) {
+      /* Compass .plt file */
       if (pimg->pending > 0) {
 	 /* -1 signals we've entered the first survey we want to
 	  * read, and need to fudge lots if the first action is 'D'...
@@ -1177,7 +1420,7 @@ img_read_item(img *pimg, img_point *p)
 	 char *line;
 	 char *q;
 	 size_t len = 0;
-	 int ch = getc(pimg->fh);
+	 int ch = GETC(pimg->fh);
 
 	 switch (ch) {
 	    case '\x1a': case EOF: /* Don't insist on ^Z at end of file */
@@ -1189,9 +1432,9 @@ img_read_item(img *pimg, img_point *p)
 skip_to_N:
 	       while (1) {
 		  do {
-		     ch = getc(pimg->fh);
+		     ch = GETC(pimg->fh);
 		  } while (ch != '\n' && ch != '\r' && ch != EOF);
-		  while (ch == '\n' || ch == '\r') ch = getc(pimg->fh);
+		  while (ch == '\n' || ch == '\r') ch = GETC(pimg->fh);
 		  if (ch == 'N') break;
 		  if (ch == '\x1a' || ch == EOF) return img_STOP;
 	       }
@@ -1228,7 +1471,7 @@ skip_to_N:
 		  if (pimg->survey) {
 		     fpos = ftell(pimg->fh) - 1;
 		     fseek(pimg->fh, pimg->start, SEEK_SET);
-		     ch = getc(pimg->fh);
+		     ch = GETC(pimg->fh);
 		     pimg->pending = 0;
 		  } else {
 		     /* If a file actually has a 'D' before any 'M', then
@@ -1242,7 +1485,7 @@ skip_to_N:
 		  img_errno = IMG_OUTOFMEMORY;
 		  return img_BAD;
 	       }
-	       /* Compass store coordinates as North, East, Up = (y,x,z)! */
+	       /* Compass stores coordinates as North, East, Up = (y,x,z)! */
 	       if (sscanf(line, "%lf%lf%lf", &p->y, &p->x, &p->z) != 3) {
 		  osfree(line);
 		  if (ferror(pimg->fh)) {
@@ -1277,6 +1520,31 @@ skip_to_N:
 	       } else {
 		   memcpy(pimg->label, q, len - 1);
 	       }
+	       q += len - 1;
+	       /* Now read LRUD.  Technically, this is optional but virtually
+		* all PLT files have it (with dummy negative values if no LRUD
+		* was measured) and some versions of Compass can't read PLT
+		* files without it!
+		*/
+	       while (*q && *q <= ' ') q++;
+	       if (*q == 'P') {
+		   if (sscanf(q + 1, "%lf%lf%lf%lf",
+			      &pimg->l, &pimg->r, &pimg->u, &pimg->d) != 4) {
+		       osfree(line);
+		       if (ferror(pimg->fh)) {
+			   img_errno = IMG_READERROR;
+		       } else {
+			   img_errno = IMG_BADFORMAT;
+		       }
+		       return img_BAD;
+		   }
+		   pimg->l *= METRES_PER_FOOT;
+		   pimg->r *= METRES_PER_FOOT;
+		   pimg->u *= METRES_PER_FOOT;
+		   pimg->d *= METRES_PER_FOOT;
+	       } else {
+		   pimg->l = pimg->r = pimg->u = pimg->d = -1;
+	       }
 	       osfree(line);
 	       pimg->flags = img_SFLAG_UNDERGROUND; /* default flags */
 	       if (fpos != -1) {
@@ -1292,7 +1560,7 @@ skip_to_N:
 	 }
       }
    } else {
-      /* version -3 or -4: CMAP .xyz file */
+      /* CMAP .xyz file */
       char *line = NULL;
       char *q;
       size_t len;
@@ -1311,7 +1579,7 @@ skip_to_N:
 	    pimg->pending = img_STOP + 4;
 	    return img_MOVE;
 	 }
-	
+
 	 pimg->pending = 0;
 
 	 if (r == img_STOP) {
@@ -1319,7 +1587,7 @@ skip_to_N:
 	    read_xyz_shot_coords(p, pimg->label_buf + 16);
 	    return img_LINE;
 	 }
-	 
+
 	 return r;
       }
 
@@ -1336,7 +1604,7 @@ skip_to_N:
       if (line[0] == '\x1a') return img_STOP;
 
       len = strlen(line);
-      if (pimg->version == -3) {
+      if (pimg->version == VERSION_CMAP_STATION) {
 	 /* station variant */
 	 if (len < 37) {
 	    osfree(line);
@@ -1349,7 +1617,7 @@ skip_to_N:
 	 *q = '\0';
 
 	 read_xyz_station_coords(p, line);
- 
+
 	 /* FIXME: look at prev for lines (line + 32, 5) */
 	 /* FIXME: duplicate stations... */
 	 return img_LABEL;
@@ -1361,16 +1629,16 @@ skip_to_N:
 	    img_errno = IMG_BADFORMAT;
 	    return img_BAD;
 	 }
- 
+
 	 memcpy(old, line, 7);
 	 q = memchr(old, ' ', 7);
 	 if (!q) q = old + 7;
-	 *q = '\0'; 
+	 *q = '\0';
 
 	 memcpy(new, line + 7, 7);
 	 q = memchr(new, ' ', 7);
 	 if (!q) q = new + 7;
-	 *q = '\0'; 
+	 *q = '\0';
 
 	 pimg->flags = img_SFLAG_UNDERGROUND;
 
@@ -1381,7 +1649,7 @@ skip_to_N:
 	    osfree(line);
 	    return img_LABEL;
 	 }
- 
+
 	 if (strcmp(old, pimg->label) == 0) {
 	    pimg->pending = img_LINE + 4;
 	    read_xyz_shot_coords(p, line);
@@ -1389,7 +1657,7 @@ skip_to_N:
 	    osfree(line);
 	    return img_LABEL;
 	 }
-	 
+
 	 pimg->pending = img_LABEL + 4;
 	 read_xyz_shot_coords(p, line);
 	 strcpy(pimg->label, new);
@@ -1415,11 +1683,11 @@ write_v3label(img *pimg, int opt, const char *s)
    SVX_ASSERT(len <= pimg->label_len);
    n = pimg->label_len - len;
    if (len == 0) {
-      if (pimg->label_len) putc(0, pimg->fh);
+      if (pimg->label_len) PUTC(0, pimg->fh);
    } else if (n <= 16) {
-      if (n) putc(n + 15, pimg->fh);
+      if (n) PUTC(n + 15, pimg->fh);
    } else if (dot == 0) {
-      if (pimg->label_len) putc(0, pimg->fh);
+      if (pimg->label_len) PUTC(0, pimg->fh);
       len = 0;
    } else {
       const char *p = pimg->label_buf + dot;
@@ -1428,23 +1696,23 @@ write_v3label(img *pimg, int opt, const char *s)
 	 if (*p++ == '.') n++;
       }
       if (n <= 14) {
-	 putc(n, pimg->fh);
+	 PUTC(n, pimg->fh);
 	 len = dot;
       } else {
-	 if (pimg->label_len) putc(0, pimg->fh);
+	 if (pimg->label_len) PUTC(0, pimg->fh);
 	 len = 0;
       }
    }
 
    n = strlen(s + len);
-   putc(opt, pimg->fh);
+   PUTC(opt, pimg->fh);
    if (n < 0xfe) {
-      putc(n, pimg->fh);
+      PUTC(n, pimg->fh);
    } else if (n < 0xffff + 0xfe) {
-      putc(0xfe, pimg->fh);
+      PUTC(0xfe, pimg->fh);
       put16((short)(n - 0xfe), pimg->fh);
    } else {
-      putc(0xff, pimg->fh);
+      PUTC(0xff, pimg->fh);
       put32(n, pimg->fh);
    }
    fwrite(s + len, n, 1, pimg->fh);
@@ -1458,29 +1726,147 @@ write_v3label(img *pimg, int opt, const char *s)
    return !ferror(pimg->fh);
 }
 
+static void
+img_write_item_date(img *pimg)
+{
+    int same, unset;
+    /* Only write dates when they've changed. */
+#if IMG_API_VERSION == 0
+    if (pimg->date1 == pimg->olddate1 && pimg->date2 == pimg->olddate2)
+	return;
+
+    same = (pimg->date1 == pimg->date2);
+    unset = (pimg->date1 == 0);
+#else /* IMG_API_VERSION == 1 */
+    if (pimg->days1 == pimg->olddays1 && pimg->days2 == pimg->olddays2)
+	return;
+
+    same = (pimg->days1 == pimg->days2);
+    unset = (pimg->days1 == -1);
+#endif
+
+    if (same) {
+	if (img_output_version < 7) {
+	    PUTC(0x20, pimg->fh);
+#if IMG_API_VERSION == 0
+	    put32(pimg->date1, pimg->fh);
+#else /* IMG_API_VERSION == 1 */
+	    put32((pimg->days1 - 25567) * 86400, pimg->fh);
+#endif
+	} else {
+	    if (unset) {
+		PUTC(0x24, pimg->fh);
+	    } else {
+		PUTC(0x20, pimg->fh);
+#if IMG_API_VERSION == 0
+		put16(pimg->date1 / 86400 + 25567, pimg->fh);
+#else /* IMG_API_VERSION == 1 */
+		put16(pimg->days1, pimg->fh);
+#endif
+	    }
+	}
+    } else {
+	if (img_output_version < 7) {
+	    PUTC(0x21, pimg->fh);
+#if IMG_API_VERSION == 0
+	    put32(pimg->date1, pimg->fh);
+	    put32(pimg->date2, pimg->fh);
+#else /* IMG_API_VERSION == 1 */
+	    put32((pimg->days1 - 25567) * 86400, pimg->fh);
+	    put32((pimg->days2 - 25567) * 86400, pimg->fh);
+#endif
+	} else {
+#if IMG_API_VERSION == 0
+	    int diff = (pimg->date2 - pimg->date1) / 86400;
+	    if (diff > 0 && diff <= 256) {
+		PUTC(0x21, pimg->fh);
+		put16(pimg->date1 / 86400 + 25567, pimg->fh);
+		PUTC(diff - 1, pimg->fh);
+	    } else {
+		PUTC(0x23, pimg->fh);
+		put16(pimg->date1 / 86400 + 25567, pimg->fh);
+		put16(pimg->date2 / 86400 + 25567, pimg->fh);
+	    }
+#else /* IMG_API_VERSION == 1 */
+	    int diff = pimg->days2 - pimg->days1;
+	    if (diff > 0 && diff <= 256) {
+		PUTC(0x21, pimg->fh);
+		put16(pimg->days1, pimg->fh);
+		PUTC(diff - 1, pimg->fh);
+	    } else {
+		PUTC(0x23, pimg->fh);
+		put16(pimg->days1, pimg->fh);
+		put16(pimg->days2, pimg->fh);
+	    }
+#endif
+	}
+    }
+#if IMG_API_VERSION == 0
+    pimg->olddate1 = pimg->date1;
+    pimg->olddate2 = pimg->date2;
+#else /* IMG_API_VERSION == 1 */
+    pimg->olddays1 = pimg->days1;
+    pimg->olddays2 = pimg->days2;
+#endif
+}
+
 void
 img_write_item(img *pimg, int code, int flags, const char *s,
 	       double x, double y, double z)
 {
    if (!pimg) return;
-   if (pimg->version == 3) {
+   if (pimg->version >= 3) {
       int opt = 0;
       switch (code) {
        case img_LABEL:
 	 write_v3label(pimg, 0x40 | flags, s);
 	 opt = 0;
 	 break;
+       case img_XSECT: {
+	 INT32_T l, r, u, d, max_dim;
+	 /* Need at least version 5 for img_XSECT. */
+	 if (pimg->version < 5) break;
+	 img_write_item_date(pimg);
+	 l = (INT32_T)my_round(pimg->l * 100.0);
+	 r = (INT32_T)my_round(pimg->r * 100.0);
+	 u = (INT32_T)my_round(pimg->u * 100.0);
+	 d = (INT32_T)my_round(pimg->d * 100.0);
+	 if (l < 0) l = -1;
+	 if (r < 0) r = -1;
+	 if (u < 0) u = -1;
+	 if (d < 0) d = -1;
+	 max_dim = max(max(l, r), max(u, d));
+	 flags &= 1;
+	 if (max_dim >= 32768) flags |= 2;
+	 write_v3label(pimg, 0x30 | flags, s);
+	 if (flags & 2) {
+	    /* Big passage!  Need to use 4 bytes. */
+	    put32(l, pimg->fh);
+	    put32(r, pimg->fh);
+	    put32(u, pimg->fh);
+	    put32(d, pimg->fh);
+	 } else {
+	    put16(l, pimg->fh);
+	    put16(r, pimg->fh);
+	    put16(u, pimg->fh);
+	    put16(d, pimg->fh);
+	 }
+	 return;
+       }
        case img_MOVE:
 	 opt = 15;
 	 break;
        case img_LINE:
+	 if (pimg->version >= 4) {
+	     img_write_item_date(pimg);
+	 }
 	 write_v3label(pimg, 0x80 | flags, s ? s : "");
 	 opt = 0;
 	 break;
        default: /* ignore for now */
 	 return;
       }
-      if (opt) putc(opt, pimg->fh);
+      if (opt) PUTC(opt, pimg->fh);
       /* Output in cm */
       put32((INT32_T)my_round(x * 100.0), pimg->fh);
       put32((INT32_T)my_round(y * 100.0), pimg->fh);
@@ -1503,12 +1889,12 @@ img_write_item(img *pimg, int code, int flags, const char *s,
 	    /* long label - not in early incarnations of v2 format, but few
 	     * 3d files will need these, so better not to force incompatibility
 	     * with a new version I think... */
-	    putc(7, pimg->fh);
-	    putc(flags, pimg->fh);
+	    PUTC(7, pimg->fh);
+	    PUTC(flags, pimg->fh);
 	    put32(len, pimg->fh);
 	    fputs(s, pimg->fh);
 	 } else {
-	    putc(0x40 | (flags & 0x3f), pimg->fh);
+	    PUTC(0x40 | (flags & 0x3f), pimg->fh);
 	    fputsnl(s, pimg->fh);
 	 }
 	 opt = 0;
@@ -1529,7 +1915,7 @@ img_write_item(img *pimg, int code, int flags, const char *s,
       if (pimg->version == 1) {
 	 put32(opt, pimg->fh);
       } else {
-	 if (opt) putc(opt, pimg->fh);
+	 if (opt) PUTC(opt, pimg->fh);
       }
       /* Output in cm */
       put32((INT32_T)my_round(x * 100.0), pimg->fh);
@@ -1538,6 +1924,25 @@ img_write_item(img *pimg, int code, int flags, const char *s,
    }
 }
 
+/* Write error information for the current traverse
+ * n_legs is the number of legs in the traverse
+ * length is the traverse length (in m)
+ * E is the ratio of the observed misclosure to the theoretical one
+ * H is the ratio of the observed horizontal misclosure to the theoretical one
+ * V is the ratio of the observed vertical misclosure to the theoretical one
+ */
+void
+img_write_errors(img *pimg, int n_legs, double length,
+		 double E, double H, double V)
+{
+    PUTC(0x22, pimg->fh);
+    put32(n_legs, pimg->fh);
+    put32((INT32_T)my_round(length * 100.0), pimg->fh);
+    put32((INT32_T)my_round(E * 100.0), pimg->fh);
+    put32((INT32_T)my_round(H * 100.0), pimg->fh);
+    put32((INT32_T)my_round(V * 100.0), pimg->fh);
+}
+
 int
 img_close(img *pimg)
 {
@@ -1555,11 +1960,11 @@ img_close(img *pimg)
 	       put32((INT32_T)-1, pimg->fh);
 	       break;
 	     case 2:
-	       putc(0, pimg->fh);
+	       PUTC(0, pimg->fh);
 	       break;
-	     case 3:
-	       if (pimg->label_len) putc(0, pimg->fh);
-	       putc(0, pimg->fh);
+	     default:
+	       if (pimg->label_len) PUTC(0, pimg->fh);
+	       PUTC(0, pimg->fh);
 	       break;
 	    }
 	 }
@@ -1568,6 +1973,9 @@ img_close(img *pimg)
 	 if (!result) img_errno = pimg->fRead ? IMG_READERROR : IMG_WRITEERROR;
       }
       osfree(pimg->label_buf);
+#ifdef IMG_HOSTED
+      osfree(pimg->filename_opened);
+#endif
       osfree(pimg);
    }
    return result;
